module 'data.fd.pac' {
   /** + A = A.
     * Identity relation.
     */

   + A = A.

   /** - A = B.
     * Ensures that finite-domain variables <code>A</code> and <code>B</code>
     * are opposite.
     */

   - A = 'kernel':fd_pac_opp(A).

   /** A + B = R.
     * Ensures that the sum of finite-domain variables <code>A</code> and <code>B</code>
     * is equal to the finite-domain variable <code>R</code>.
     */

   A + B = 'kernel':fd_pac_add(A, B).

   /** A - B = R.
     * Ensures that the difference between finite-domain variables <code>A</code> and <code>B</code>
     * is equal to the finite-domain variable <code>R</code>.
     */

   A - B = 'kernel':fd_pac_sub(A, B).

   /** A * B = R.
     * Ensures that the product of finite-domain variables <code>A</code> and <code>B</code>
     * is equal to the finite-domain variable <code>R</code>.
     */

   A * B = 'kernel':fd_pac_mul(A, B).

   /** A / B = R.
     * Ensures that the division of finite-domain variables <code>A</code> and <code>B</code>
     * is equal to the finite-domain variable <code>R</code>.
     */

   A / B = 'kernel':fd_pac_div(A, B).

   /** (A = B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  = <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   (A = B) =
      'kernel':fd_pac_eq(A, B).

   /** A < B.
    *  Succeeds if <code>A</code>
    *  &lt; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   A < B :-
      'kernel':fd_pac_lt(A, B).

   /** (A < B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &lt; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   (A < B) =
      'kernel':fd_pac_lt(A, B).

   /** A =< B.
    *  Succeeds if <code>A</code>
    *  &le; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   A =< B :-
      'kernel':fd_pac_le(A, B).

   /** (A =< B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &le; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   (A =< B) =
      'kernel':fd_pac_le(A, B).

   /** A > B.
    *  Succeeds if <code>A</code>
    *  &gt; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   A > B :-
      'kernel':fd_pac_gt(A, B).

   /** (A > B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &gt; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   (A > B) =
      'kernel':fd_pac_gt(A, B).

   /** A >= B.
    *  Succeeds if <code>A</code>
    *  &ge; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   A >= B :-
      'kernel':fd_pac_ge(A, B).

   /** (A >= B) = X.
    *  Succeeds if <code>X</code> is the reified constraint <code>A</code>
    *  &ge; <code>B</code>, where <code>A</code> and <code>B</code> are
    *  finite-domain variables.
    */

   (A >= B) =
      'kernel':fd_pac_ge(A, B).
}