module 'data.module' {
   /** check(X).
    *  Succeeds if <code>X</code> is instantiated to a module.
    */

   check(Module) :-
      'data.term':check(Module),
      'data.term':decomposition(Module, [F, _]),
      'data.atom':sub(F, 0, _, _, '<module>:')
.
   /** predicate(Module, F, N).
    *  <prerequisite>
    *  <code>Module</code> is an instantiated module.
    *  </prerequisite>
    *  <prerequisite>
    *  <code>F</code> is an instantiated atom.
    *  </prerequisite>
    *  <prerequisite>
    *  <code>N</code> is an instantiated integer.
    *  </prerequisite>
    *  Succeeds if <code>Module</code> has a predicate with signature <code>F/N</code>.
    */

   predicate(Module, F, N) :-
      (
         'data.atom':check(F),
         'data.number':check(N)
      ->
         Module:'$has_predicate'('data.term':new(F, N))
      ;
         Module:'$has_predicate'(P),
         'data.term':functor(P, F, N)
      ).

   /** predicate(Module, F/N).
    *  Equivalent to <code>predicate(Module, F, N)</code>.
    */

   predicate(Module, F/N) :-
      'data.module':predicate(Module, F, N).
}