module 'emop.emop' [Impl] { 
   main :-
      Impl:main.

   lookup(ModSpec, Pos, BaseFilename, FilenameMpl, SMpl) :-
      Impl:lookup(ModSpec, Pos, BaseFilename, FilenameMpl, SMpl).
},

module Impl [Impl, Paths, Action, Output, Modules] {
   main :-
      catch(
         (
            Impl:paths(Paths),
            'sys.cmdline':list(Arguments),
            (
               Arguments = []
            ->
               Action = 'top'
            ;
               Impl:parse_arguments(Arguments)
            ),
            (
               Action = 'build'
            ->
               'emop.compile':build(Modules, Output)
            ;
               Action = 'test'
            ->
               'emop.compile':test(Modules, Output)
            ;
               Action = 'top'
            ->
               'emop.top':main
            ;
               Action = 'version'
            ->
               Impl:version
            ;
               Action = 'help'
            ->
               Impl:help
            ),
            'sys.process':halt
         ),
         E,
         (
            'emop.message':error(E),
            'sys.process':abort
         )
      ).

   version :-
      'io.std':(
         print_endline('EMoP version 0.01'),
         print_endline('Copyright (C) 2008 INRIA, Institut National de Recherche en'),
         print_endline('Informatique et en Automatique, France.'),
         endline,
         print_endline('This is free software.  You may redistribute copies of it under'),
         print_endline('the terms of the GNU General Public Licence'),
         print_endline('<http://www.gnu.org/licences/gpl.html>.'),
         endline,
         print_endline('There is NO WARRANTY, to the extent permitted by law.'),
         endline,
         print_endline('Written by Thierry Martinez <Thierry.Martinez@inria.fr>.')
      ).

   help :-
      ProgramName = 'sys.cmdline':{0},
      'io.std':(
         print_endline('Usage:\n'),
         format('~a\n\tRun top-level.\n\n', [ProgramName]),
         format(
            '~a [build] <main module> [<other module>]...\n\tBuild given modules.\n\n',
            [ProgramName]),
         format(
            '~a test <module>...\n\tTest given modules.\n\n',
            [ProgramName]),
         format(
            '~a version\n\tPrint version informations and exit.\n\n',
            [ProgramName]),
         format(
            '~a help\n\tPrint this help and exit.\n\n',
            [ProgramName]),
         print_endline('Other options:'),
         print_endline('--output-filename=FILENAME \tSet output filename.'),
         endline,
         print_endline('Options are prefixed with `-\' or `--\' and can be abbreviated'),
         print_endline('to any non-ambiguous prefix.  Actions can be prefixed and then'),
         print_endline('abbreviated too.  Version informations and help are available'),
         print_endline('through usual options `-h\' and `-t\' respectively.')
      ).

   parse_arguments(Arguments) :-
      RevModulesRef = 'data.ref.backtrackable':new([]),
      module Aux [Aux, RevModulesRef, Action, Output] {
         action(build).
         action(test).
         action(top).
         action(help).
         action(version).

         option(Option, M) :-
            Aux:action(Option),
            module M [Option, Aux] {
               do(Arguments, Arguments) :-
                  Aux:set_action(Option).
            }.
         option('output-filename', M) :-
            module M [Aux] {
               do([], _) :-
                  throw(error('Missing output filename', cmdline)).
                  
               do([Output | Arguments], Arguments) :-
                  Aux:set_output(Output).
            }.

         do([]).
         do([Argument | Arguments]) :-
            (
               'data.var':check(Action),
               Aux:action(Argument)
            ->
               Aux:set_action(Argument),
               Tail = Arguments
            ;
               Argument = '-h'
            ->
               Aux:set_action(help),
               Tail = Arguments
            ;
               Argument = '-v'
            ->
               Aux:set_action(version),
               Tail = Arguments
            ;
               'data.atom':sub(Argument, 0, 1, _, '-'),
               (
                  'data.atom':sub(Argument, 1, 1, _, '-')
               ->
                  'data.atom':sub(Argument, 2, _, 0, Option0)
               ;
                  'data.atom':sub(Argument, 1, _, 0, Option0)
               )
            ->
               (
                  'data.atom':sub(Option0, Left, _, Right, '=')
               ->
                  'data.atom':sub(Option0, 0, Left, _, Option),
                  'data.atom':sub(Option0, _, Right, 0, OptionArg),
                  OptionArgs = [OptionArg],
                  OptionTail = [],
                  Arguments = Tail
               ;
                  Option = Option0,
                  OptionArgs = Arguments,
                  OptionTail = Tail
               ),
               (
                  Option = ''
               ->
                  throw(
                     error(
                        'Empty option name',
                        cmdline
                     )
                  )
               ;
                  true
               ),
               'data.atom':string(Option, OptionStr),
               'data.list':append(OptionStr, _, OptionStrPartial),
               findall(
                  OptionFullname,
                  (
                     Aux:option(OptionFullname, _),
                     'data.atom':string(OptionFullname, OptionStrPartial)
                  ),
                  Options
               ),
               (
                  Options = [OptionFullname]
               ->
                  Aux:option(OptionFullname, M)
               ;
                  Options = []
               ->
                  throw(
                     error(
                        'data.atom':format('Unknown option \'~a\'', [Option]),
                        cmdline
                     )
                  )
               ;
                  Options = [_ | _]
               ->
                  throw(
                     error(
                        'data.atom':format('Ambiguous option \'~a\'', [Option]),
                        cmdline
                     )
                  )
               ),    
               M:do(OptionArgs, Tail0),
               (
                  OptionTail = Tail0
               ->
                  true
               ;
                  throw(
                     error(
                        'data.atom':format(
                           'No argument for ~a',
                           [OptionFullname]
                        ),
                        cmdline
                     )
                  )
               )
            ;
               Aux:queue_module(Argument),
               Tail = Arguments
            ),
            Aux:do(Tail).

         set_action(New) :-
            (
               'data.var':check(Action)
            ->
               Action = New
            ;
               throw(error('There cannot be more than one action', cmdline))
            ).

         set_output(New) :-
            (
               'data.var':check(Output)
            ->
               Output = New
            ;
               throw(
                  error(
                     'There cannot be more than one output filename',
                     cmdline
                  )
               )
            ).

         queue_module(Name) :-
            (
               'data.atom':sub(Name, _, 4, 0, '.mpl')
            ->
               Module = file(Name)
            ;
               Module = name(Name, [], '.')
            ),
            RevModulesRef:set([Module | RevModulesRef:get]).
      },
      'data.list':(
         Aux:do(Arguments),
         reverse(RevModulesRef:get, Modules)
      ).

   lookup(ModSpec, Pos, BaseFilename, FilenameMpl, SMpl) :-
      (
         (
            ModSpec = file(ModName)
         ->
            FilenameMpl = 'sys.file':real(ModName),
            'data.atom':sub(FilenameMpl, 0, _, 4, BaseFilename)
         ;
            ModSpec = name(ModName, PackageNodes, RootPackagePath)
         ->
            Impl:module_filename(ModName, Pos, RelBaseFilename),
            (
               'data.list':longuest_prefix(Prefix, PackageNodes),
               'data.path':(
                  join(Prefix, RelPath),
                  concat(RootPackagePath, RelPath, Path),
                  concat(Path, RelBaseFilename, BaseFilename)
               )
            ;
               'data.list':member(Path, Paths),
               'data.path':concat(Path, RelBaseFilename, BaseFilename)
            )
         ),
         'data.atom':concat(BaseFilename, '.mpl', FilenameMpl),
         catch(
            'io.file':open_read(FilenameMpl, SMpl),
            error(existence_error(_, _), _),
            fail
         )
      ->
         true
      ;
         (
            ModSpec = file(ModName)
         ->
            true
         ;
            ModSpec = name(ModName, PackageNodes, RootPackagePath)
         ),
         'data.atom':format('Unable to find \'~a\'', [ModName], Message),
         throw(error(Message, Pos))
      ).

   module_filename(Mod, Pos, Filename) :-
      'data.atom':string(Mod, Str),
      module Change {
         do('.', '/') :- !.
         do('/', _) :- fail.
         do(C, C).
      },
      (
         'data.list':iter(Str, Change, FilenameStr)
      ->
         'data.string':atom(FilenameStr, Filename)
      ;
         'data.atom':format('Invalid module name \'~a\'', [Mod], Message),
         throw(error(Message, Pos))
      ).

   lib_dir = 'data.path':concat('data.path':base(ExecDir), 'lib') :-
      ExecDir = 'data.path':base('sys.file':real('sys.process':exec_filename)).

   paths(Paths) :-
      (
         Path = 'sys.env':{'EMOPPATH'}
      ->
         'data.atom':split(Path, ':', EnvPaths)
      ;
         EnvPaths = []
      ),
      'data.list':append(EnvPaths, [Impl:lib_dir], Paths).
}
