module 'emop.expand' [Impl] {
   expand_file(Fid, FileXml, Ast, Clauses, Dependencies, Goal, Expander) :-
      'data.ref.backtrackable':(
         new([], DependenciesRef),
         new([], ClausesRef),
         new(0, PredicateIdCounterRef),
         new(0, FreshModuleCounterRef),
         new(0, FreshVarCounterRef)
      ),
      module ExpandFile [
         ExpandFile,
         Impl,
         Fid,
         FileXml,
         DependenciesRef,
         ClausesRef,
         PredicateIdCounterRef,
         FreshModuleCounterRef,
         FreshVarCounterRef
      ] {
         predicate_id(PredicateId) :-

            PredicateId = 'data.atom':format(
               'p~d', ['data.number':(PredicateIdCounterRef++)]
            ).

         add_clause(F, N, Args, B) :-
            ClausesRef:get(Clauses),
            ClausesRef:set([clause(F, N, Args, B) | Clauses]).

         add_dependency(Mod, Pos) :-
            DependenciesRef:get(Dependencies),
            (
               'data.list':member(dep(Pos, _, _), Dependencies)
            ->
               true
            ;
               DependenciesRef:set([dep(Mod, Pos) | Dependencies])
            ).

         fresh_module_index = 'data.number':(FreshModuleCounterRef++).

         fresh_var(V) :-
            V = var(
               'data.atom':format(
                  '_~d', ['data.number':(FreshVarCounterRef++)]
               )
            ).

         expand_module(Name, Env, Body, Sup, Expander, Goal) :-
            EnvRec = 'data.record':{},
            Impl:extract_env(Env, Expander, EnvRec, EnvList0),
            (
               Sup = none
            ->
               ScopeStack = [],
               EnvList = EnvList0
            ;
               Sup = some(_, SupId, _, SupEnvList, SupSgn, SupScopeStack)
            ->
               EnvList = 'data.list':iter_fold(
                  SupEnvList,
                  SupEnvList0,
                  EnvList0
               ) [ExpandFile] {
                  do(Var0, Var1, EnvList0, EnvList1) :-
                     (
                        'data.list':member(Var0, EnvList0)
                     ->
                        Var1 = Var0,
                        EnvList1 = EnvList0
                     ;
                        ExpandFile:fresh_var(Var1),
                        EnvList1 = [Var1 | EnvList0]
                     ).
               },
               ParentScope = scope(SupId, SupEnvList0, SupSgn),
               ScopeStack = [ParentScope | SupScopeStack]
            ),
            This = some(ModTerm, Id, EnvTerm, EnvList, Sgn, ScopeStack),
            module [
               Title,
               Goal,
               Fid,
               Id,
               ModTerm,
               ExpandFile,
               Expander,
               EnvList,
               EnvTerm
            ] {
               expand_module_name(var(Title, Pb, Pe)) :-
                  ExpandFile:fresh_module_index(Index),
                  ExpandFile:fresh_var(F),
                  (
                     Fid = toplevel(Fid0)
                  ->
                     EmopCall = 'dynamic_emop_call'
                  ;
                     EmopCall = 'static_emop_call',
                     Fid0 = Fid
                  ),
                  'data.atom':format('<module>:~a@~d', [Fid0, Index], Id),
                  EnvTerm = callable('env', EnvList),
                  Expander:var_lookup(Title, Pb, Pe, _),
                  ExpandFile:add_clause(
                     EmopCall,
                     3,
                     [callable(Id, [var(env)]), F, var(pos)],
                     callable(
                        ',',
                        [
                           callable('!', []),
                           callable(Id, [F, var(env), var(pos)])
                        ]
                     )
                  ),
                  ModTerm = callable(Id, [EnvTerm]),
                  Goal = callable('=', [var(Title), ModTerm]).
      
               expand_module_name(atom(Id, _, _)) :-
                  (
                     Fid = toplevel(Fid0)
                  ->
                     EmopCall = 'dynamic_emop_call'
                  ;
                     EmopCall = 'static_emop_call',
                     Fid0 = Fid
                  ),
                  ExpandFile:fresh_var(F),
                  EnvTerm = callable('env', EnvList),
                  Title = Id,
                  ModTerm = callable(Id, []),
                  ExpandFile:add_clause(
                     EmopCall,
                     3,
                     [ModTerm, F, var(pos)],
                     callable(
                        ',',
                        [
                           callable('!', []),
                           callable(
                              ',',
                              [
                                 callable('g_read', [ModTerm, EnvTerm]),
                                 callable(Id, [F, EnvTerm, var(pos)])
                              ]
                           )
                        ]
                     )
                  ),
                  Goal = callable('g_link', [ModTerm, EnvTerm]).
            }:expand_module_name(Name),
            'data.atom':format('~a:$has_predicate', [Id], HasPredicate),
            ExpandFile:add_clause(
               Id,
               3,
               [callable('$has_predicate', [var(pred)]), var('_'), var('_')],
               callable(
                  ',',
                  [
                     callable('!', []),
                     callable(HasPredicate, [var(pred)])
                  ]
               )
            ),
            FileXml:format('<module title="~a">\n', [Title]),
            NewClausesRef = 'data.ref.backtrackable':new([]),
            module EnumClauses [
               ExpandFile,
               EnumClauses,
               NewClausesRef
            ] {
               expand_clauses(callable(F, _, Args, Pb, Pe)) :-
                  EnumClauses:expand_clauses_callable(F, Args, Pb, Pe, nodoc).
               expand_clauses(reserved(F, Args, Pb, Pe)) :-
                  EnumClauses:expand_clauses_reserved(F, Args, Pb, Pe, nodoc).
               expand_clauses('()'(Sub_clauses, _, _)) :-
                  EnumClauses:expand_clauses(Sub_clauses).
               expand_clauses('module'(_, _, _, Pb, Pe)) :-
                  throw(error('A module is not a valid clause', file(Pb, Pe))).
               expand_clauses(number(_, Pb, Pe)) :-
                  throw(error('A number is not a valid clause', file(Pb, Pe))).
               expand_clauses(var(_, Pb, Pe)) :-
                  throw(error('A variable is not a valid clause', file(Pb, Pe))).
               expand_clauses(empty(_, _)).
               expand_clauses(doc(Comment, Pb, Pe, Clause)) :-
                  EnumClauses:expand_clause_doc(Clause, doc(Comment, Pb, Pe)).
               
               expand_clause_doc(callable(F, _, Args, Pb, Pe), Doc) :-
                  EnumClauses:expand_clauses_callable(F, Args, Pb, Pe, Doc).
               expand_clause_doc(reserved(F, Args, Pb, Pe), Doc) :-
                  EnumClauses:expand_clauses_reserved(F, Args, Pb, Pe, Doc).
               expand_clause_doc('()'(Sub_clauses, _, _), Doc) :-
                  EnumClauses:expand_clause_doc(Sub_clauses, Doc).
               expand_clause_doc('module'(_, _, _, Pb, Pe), _) :-
                  throw(error('A module is not a valid clause', file(Pb, Pe))).
               expand_clause_doc(number(_, Pb, Pe), _) :-
                  throw(error('A number is not a valid clause', file(Pb, Pe))).
               expand_clause_doc(var(_, Pb, Pe), _) :-
                  throw(error('A variable is not a valid clause', file(Pb, Pe))).
               expand_clause_doc(empty(Pb, Pe), _) :-
                  throw(error('A documentation should be followed by an implementation', file(Pb, Pe))).
               expand_clause_doc(doc(_, Pb, Pe, _), _) :-
                  throw(error('Predicate already documented', file(Pb, Pe))).

               expand_xml(Atom, Xml) :-
                  'data.atom':subst(
                     Atom,
                     ['<' => '&lt;', '>' => '&gt;', '\"' => '&quot;', '\'' => '&apos;'],
                     Xml
                  ).
      
               expand_clauses_reserved(':-', [H, B], _, _, Doc) :-
                  !,
                  EnumClauses:(
                     decompose_head(H, Pb, Pe, Status, F, Args),
                     (
                        Status = reserved
                     ->
                        (
                           F = '=',
                           Args = [A, R]
                        ->
                           EnumClauses:expand_eq_clause(A, R, Pb, Pe, B, Doc)
                        ;
                           throw(
                              error(
                                 'A predicate cannot overload a reserved '
                                 'construction',
                                 file(Pb, Pe)
                              )
                           )
                        )
                     ;
                        NewClausesRef:set(
                           [clause(F, Args, B, Doc) | NewClausesRef:get]
                        )
                     )
                  ).
               expand_clauses_reserved('=', [A, R], Pb, Pe, Doc) :-
                  !,
                  EnumClauses:expand_eq_clause(
                     A, R, Pb, Pe, reserved(true, [], Pb, Pe), Doc
                  ).
               expand_clauses_reserved(_, _, Pb, Pe, _) :-
                  throw(
                     error(
                        'A predicate cannot overload a reserved construction',
                        file(Pb, Pe)
                     )
                  ).
      
               expand_clauses_callable('.', [A0, A1], Pb, Pe, Doc) :-
                  !,
                  (
                     Doc = nodoc
                  ->
                     EnumClauses:(
                        expand_clauses(A0),
                        expand_clauses(A1)
                     )
                  ;
                     throw(
                        error(
                           'A documentation should be followed by an implementation',
                           file(Pb, Pe)
                        )
                     )
                  ).
               expand_clauses_callable(F, Args, Pb, Pe, Doc) :-
                  NewClausesRef:set(
                     [clause(F, Args, reserved(true, [], Pb, Pe), Doc)
                     | NewClausesRef:get]
                  ).
      
               expand_eq_clause(A, R, Pb, Pe, B, Doc) :-
                  module Aux [Aux, F, Args] {
                     do(callable(F, _, Args, _, _)).
                     do('()'(Sub_clauses, _, _)) :-
                        Aux:do(Sub_clauses).
                     do(reserved(':', [_, _], Pb, Pe)) :-
                        throw(
                           error(
                              'Predicate definition cannot be prefixed',
                              file(Pb, Pe)
                           )
                        ).
                     do(reserved(F, Args, _, _)).
                     do('module'(_, _, _, Pb, Pe)) :-
                        throw(
                           error(
                              'A module is not a valid predicate',
                              file(Pb, Pe)
                           )
                        ).
                     do(number(_, Pb, Pe)) :-
                        throw(
                           error(
                              'A number is not a valid predicate',
                              file(Pb, Pe)
                           )
                        ).
                     do(var(_, Pb, Pe)) :-
                        throw(
                           error(
                              'A variable is not a valid predicate',
                              file(Pb, Pe)
                           )
                        ).
                     do(empty(Pb, Pe)) :-
                        throw(
                           error(
                              'An empty item is not a valid predicate',
                              file(Pb, Pe))
                        ).
                     do(doc(_, Pb, Pe, _)) :-
                        throw(
                           error('Unallowed documentation here', file(Pb, Pe))
                        ).
                  },
                  Aux:do(A),
                  ExpandFile:fresh_var(V),
                  V = var(FreshVarName),
                  V0 = var(FreshVarName, Pb, Pe),
                  B0 = reserved(',', [B, reserved('=', [V0, R], Pb, Pe)], Pb, Pe),
                  'data.list':append(Args, [V0], Args0),
                  NewClausesRef:set([clause(F, Args0, B0, Doc) | NewClausesRef:get]).
      
               decompose_head(
                  callable(F, _, Args, Pb, Pe), Pb, Pe, callable, F, Args
               ).
               decompose_head(
                  reserved(F, Args, Pb, Pe), Pb, Pe, reserved, F, Args
               ).
               decompose_head('()'(SubHead, _, _), Pb, Pe, Status, F, Args) :-
                  EnumClauses:decompose_head(SubHead, Pb, Pe, Status, F, Args).
               decompose_head('module'(_, _, _, Pb, Pe), _, _, _, _, _) :-
                  throw(error('A module is not a valid head', file(Pb, Pe))).
               decompose_head(number(_, Pb, Pe), _, _, _, _, _) :-
                  throw(error('A number is not a valid head', file(Pb, Pe))).
               decompose_head(var(_, Pb, Pe), _, _, _, _, _) :-
                  throw(error('A variable is not a valid head', file(Pb, Pe))).
               decompose_head(empty(Pb, Pe), _, _, _, _, _) :-
                  throw(error('Empty head unallowed', file(Pb, Pe))).
               decompose_head(doc(_, Pb, Pe, _), _, _, _, _, _) :-
                  throw(
                     error('Unallowed documentation here', file(Pb, Pe))
                  ).
            }:expand_clauses(Body),
            NewClauses = 'data.list':reverse(NewClausesRef:get),            
            Sgn = 'data.list':fold_left(NewClauses, []) [
               Id,
               HasPredicate,
               ExpandFile
            ] {
               proto_args(N, Args) :-
                  (
                     'data.number':(N > 0)
                  ->
                     'data.atom':format('__~d', [N], Atom),
                     Args = [var(Atom) | Tail],
                     'data.number':pred(N, M),
                     this:proto_args(M, Tail)
                  ;
                     Args = []
                  ).

               do(clause(F, Args, _, Doc), Sgn, NewSgn) :-
                  'data.list':length(Args, N),
                  (
                     'data.list':member(F / N, Sgn)
                  ->
                     (
                        Doc = nodoc
                     ->
                        NewSgn = Sgn
                     ;
                        Doc = doc(_, Pb, Pe)
                     ->
                        throw(
                           error(
                              'Documentation should precede the first clause of a predicate',
                              file(Pb, Pe)
                           )
                        )
                     )
                  ;
                     this:proto_args(N, ProtoArgs),
                     'data.list':append(
                        ProtoArgs,
                        [var(env)],
                        ProtoEnv
                     ),
                     ExpandFile:add_clause(
                        HasPredicate,
                        1,
                        [callable(F, ProtoArgs)],
                        callable('true', [])
                     ),
                     'data.atom':format('~a:~a', [Id, F], PrefixedF),
                     ExpandFile:add_clause(
                        Id,
                        3,
                        [callable(F, ProtoArgs), var(env), var('_')],
                        callable(
                           ',',
                           [
                              callable('!', []),
                              callable(PrefixedF, ProtoEnv)
                           ]
                        )
                     ),
                     'data.atom':format('~a:~a/~d@help', [Id, F, N], PrefixedDoc),
                     (
                        Doc = nodoc
                     ->
                        true
                     ;
                        Doc = doc(Comment, _, _)
                     ->
                        ExpandFile:add_clause(
                           PrefixedDoc,
                           1,
                           [var(help)],
                           callable(
                              '=',
                              [
                                 var(help),
                                 callable(Comment, [])
                              ]
                           )
                        )
                     ),
                     NewSgn = [F / N | Sgn]
                  ).
            },
            'data.list':iter(NewClauses) [
               Id,
               EnvTerm,
               EnvRec,
               This,
               ExpandFile
            ] {
               do(clause(F, Args, B, _)) :-
                  LocalExpander = ExpandFile:expander(EnvRec:copy, This),
                  LocalExpander:expand_goal(B, none, Body1),
                  'data.atom':format('~a:~a', [Id, F], PrefixedF),
                  Args1 = 'data.list':iter(Args) [LocalExpander] {
                     do(callable(F, _, Args, Pb, Pe), Result) :-
                        this:do_callable(F, Args, Pb, Pe, Result).
                     do(reserved(':', [_, _], Pb, Pe), _) :-
                        !,
                        throw(
                           error('Arguments cannot be prefixed', file(Pb, Pe))
                        ).
                     do(reserved(F, Args, Pb, Pe), Result) :-
                        this:do_callable(F, Args, Pb, Pe, Result).
                     do('()'(SubTerm, _, _), Result) :-
                        this:do(SubTerm, Result).
                     do('module'(_, _, _, Pb, Pe), _) :-
                        throw(
                           error('A module is not a valid term', file(Pb, Pe))
                        ).
                     do(number(Number, _, _), Result) :-
                        Result = number(Number).
                     do(var(Var, Pb, Pe), Result) :-
                        LocalExpander:var_lookup(Var, Pb, Pe, _),
                        Result = var(Var).
                     do(empty(Pb, Pe), _) :-
                        throw(
                           error('Empty term unallowed', file(Pb, Pe))
                        ).
                     do(doc(_, Pb, Pe, _), _) :-
                        throw(
                           error('Unallowed documentation here', file(Pb, Pe))
                        ).
      
                     do_callable(F, Args0, _, _, Result) :-
                        Result = callable(F, Args1),
                        'data.list':iter(Args0, this, Args1).
                  },
                  'data.list':(
                     append(Args1, [EnvTerm], ArgsEnv),
                     length(Args1, N)
                  ),
                  LocalExpander:singleton_warning,
                  ExpandFile:add_clause(PrefixedF, N, ArgsEnv, Body1).
            },
            FileXml:print_endline('</module>'),
            ExpandFile:add_clause(
               Id,
               3,
               [var(t), var('_'), var(pos)],
               callable(
                  ',',
                  [
                     callable('functor', [var(t), var(f), var(n)]),
                     callable(
                        throw,
                        [
                           callable(
                              error,
                              [
                                 callable(
                                    existence_error,
                                    [
                                       callable(procedure, []),
                                       callable('/', [var(f), var(n)])
                                    ]
                                 ),
                                 var(pos)
                              ]
                           )
                        ]
                     )
                  ]
               )
            ),
            module Warning {
               do(Var, var(PB, PE, _, Status)) :-
                  (
                     Status = singleton
                  ->
                     'emop.message':warning(
                        'data.atom':format(
                           'Environment variable never used \'~a\'',
                           [Var]
                        ),
                        file(PB, PE)
                     )
                  ;
                     true
                  ).
            },
            EnvRec:iter(Warning).

         expander(Vars, This, Expander) :-
            module Expander [Vars, This, ExpandFile, Expander, Impl] {
               var_lookup(Name, PB, PE, Var) :-
                  Var = Vars:{Name},
                  (
                     'data.var':check(Var)
                  ->
                     Var = var(PB, PE, local, _)
                  ;
                     Var = var(_, _, _, non_singleton)
                  ).
      
               singleton_warning :-
                  module Warning {
                     do(Var, var(PB, PE, Kind, Status)) :-
                        (
                           Kind = local,
                           Status = singleton,
                           \+ 'data.atom':sub(Var, 0, 1, _, '_')
                        ->
                           'emop.message':warning(
                              'data.atom':format(
                                 'Singleton variable \'~a\'',
                                 [Var]
                              ),
                              file(PB, PE)
                           )
                        ;
                           true
                        ).
                  },
                  Vars:iter(Warning).
      
               expand_goal(callable(F, _, Args0, Pb, Pe), Ctx, Goal) :-
                  Expander:(
                     expand_term_list(Args0, Ctx, G0, Args1),
                     expand_call(Ctx, F, Args1, Pb, Pe, G1)
                  ),
                  Impl:seq(G0, G1, Goal).
               expand_goal(reserved(F, Args, Pb, Pe), Ctx, Goal) :-
                  Expander:expand_reserved(F, Args, Pb, Pe, Ctx, Goal).
               expand_goal('()'(SubGoal, _, _), Ctx, Goal) :-
                  Expander:expand_goal(SubGoal, Ctx, Goal).
               expand_goal('module'(Name, Env, Body, _, _), _, Goal) :-
                  ExpandFile:expand_module(
                     Name,
                     Env,
                     Body,
                     This,
                     Expander,
                     Goal
                  ).
               expand_goal(number(_, Pb, Pe), _, _) :-
                  throw(error('A number is not a valid goal', file(Pb, Pe))).
               expand_goal(var(_, Pb, Pe), _, _) :-
                  throw(error('A variable is not a valid goal', file(Pb, Pe))).
               expand_goal(empty(Pb, Pe), _, _) :-
                  throw(error('Empty goal unallowed', file(Pb, Pe))).
               expand_goal(doc(_, Pb, Pe, _), _, _) :-
                  throw(
                     error('Unallowed documentation here', file(Pb, Pe))
                  ).
      
               expand_reserved(',', [A0, B0], _, _, Ctx, Goal) :-
                  !,
                  Expander:(
                     expand_goal(A0, Ctx, A1),
                     expand_goal(B0, Ctx, B1)
                  ),
                  Goal = callable(',', [A1, B1]).
               expand_reserved('->', [A0, B0], _, _, Ctx, Goal) :-
                  !,
                  Expander:(
                     expand_goal(A0, Ctx, A1),
                     expand_goal(B0, Ctx, B1)
                  ),
                  Goal = callable('->', [A1, B1]).
               expand_reserved(';', [A0, B0], _, _, Ctx, Goal) :-
                  !,
                  Expander:(
                     expand_goal(A0, Ctx, A1),
                     expand_goal(B0, Ctx, B1)
                  ),
                  Goal = callable(';', [A1, B1]).
               expand_reserved(':', [A0, B0], Pb, Pe, Ctx, Goal) :-
                  Expander:(
                     expand_term(A0, Ctx, AG, A1),
                     expand_goal(B0, some(Used, A1), B1)
                  ),
                  Impl:check_prefix_used(Used, Pb, Pe),
                  Impl:seq(AG, B1, Goal).
               expand_reserved('\\+', [A0], _, _, Ctx, Goal) :-
                  Expander:expand_goal(A0, Ctx, A1),
                  Goal = callable('\\+', [A1]).
               expand_reserved('catch', [A0, B0, C0], _, _, Ctx, Goal) :-
                  Expander:(
                     expand_goal(A0, Ctx, A1),
                     expand_term(B0, Ctx, BG, B1),
                     expand_goal(C0, Ctx, C1)
                  ),
                  Impl:seq(BG, callable(catch, [A1, B1, C1]), Goal).
               expand_reserved('findall', [A0, B0, C0], _, _, Ctx, Goal) :-
                  Expander:(
                     expand_term(A0, Ctx, AG, A1),
                     expand_goal(B0, Ctx, B1),
                     expand_term(C0, Ctx, CG, C1)
                  ),
                  Impl:seq(
                     CG, callable(findall, [A1, Impl:seq(B1, AG), C1]), Goal
                  ).
               expand_reserved('bagof', [A0, B0, C0], _, _, Ctx, Goal) :-
                  Expander:(
                     expand_term(A0, Ctx, AG, A1),
                     expand_goal(B0, Ctx, B1),
                     expand_term(C0, Ctx, CG, C1)
                  ),
                  Impl:seq(
                     CG, callable(bagof, [A1, Impl:seq(B1, AG), C1]), Goal
                  ).
               expand_reserved('setof', [A0, B0, C0], _, _, Ctx, Goal) :-
                  Expander:(
                     expand_term(A0, Ctx, AG, A1),
                     expand_goal(B0, Ctx, B1),
                     expand_term(C0, Ctx, CG, C1)
                  ),
                  Impl:seq(
                     CG, callable(setof, [A1, Impl:seq(B1, AG), C1]), Goal
                  ).
               expand_reserved('!', [], _, _, _, Goal) :-
                  Goal = callable(!, []).
               expand_reserved('true', [], _, _, _, Goal) :-
                  Goal = callable(true, []).
               expand_reserved('fail', [], _, _, _, Goal) :-
                  Goal = callable(fail, []).
               expand_reserved('repeat', [], _, _, _, Goal) :-
                  Goal = callable(repeat, []).
               expand_reserved('for', [I0, A0, B0], _, _, Ctx, Goal) :-
                  Expander:(
                     expand_term(I0, Ctx, IG, I1),
                     expand_term(A0, Ctx, AG, A1),
                     expand_term(B0, Ctx, BG, B1)
                  ),
                  Impl:(
                     seq(IG, AG, SubGoal0),
                     seq(SubGoal0, BG, SubGoal1),
                     seq(SubGoal1, callable(for, [I1, A1, B1]), Goal)
                  ).
               expand_reserved('throw', [A0], _, _, Ctx, Goal) :-
                  Expander:expand_term(A0, Ctx, AG, A1),
                  Impl:seq(AG, callable(throw, [A1]), Goal).
               expand_reserved('=', [A0, B0], _, _, Ctx, Goal) :-
                  Expander:(
                     expand_term(A0, Ctx, AG, A1),
                     expand_term(B0, Ctx, BG, B1)
                  ),
                  Impl:(
                     seq(AG, BG, SubGoal),
                     seq(SubGoal, callable('=', [A1, B1]), Goal)
                  ).
               expand_reserved('\\=', [A0, B0], _, _, Ctx, Goal) :-
                  Expander:(
                     expand_term(A0, Ctx, AG, A1),
                     expand_term(B0, Ctx, BG, B1)
                  ),
                  Impl:(
                     seq(AG, BG, SubG),
                     seq(SubG, callable('\\=', [A1, B1]), Goal)
                  ).
      
               expand_call(none, F, Args, Pb, Pe, Goal) :-
                  (
                     This = some(_, Id, _, EnvList, Sgn, ScopeStack)
                  ->
                     'data.list':length(Args, N),
                     (
                        'data.list':member(F / N, Sgn)
                     ->
                        GoalId = Id,
                        GoalEnv = EnvList
                     ;
                        module [F, N, Pb, Pe, GoalId, GoalEnv] {
                           find([]) :-
                              throw(
                                 error(
                                    'data.atom':format(
                                       'Unknown predicate ~a/~d',
                                       [F, N]
                                    ),
                                    file(Pb, Pe)
                                 )
                              ).
                           find([scope(Id, EnvList, Sgn) | ScopeStack]) :-
                              (
                                 'data.list':member(F / N, Sgn)
                              ->
                                 GoalId = Id,
                                 GoalEnv = EnvList
                              ;
                                 this:find(ScopeStack)
                              ).
                        }:find(ScopeStack)
                     ),
                     F0 = 'data.atom':format('~a:~a', [GoalId, F]),
                     'data.list':append(Args, [callable(env, GoalEnv)], Args0),
                     Goal = callable(F0, Args0)
                  ;
                     throw(
                        error(
                           'A call outside a module should be prefixed',
                           file(Pb, Pe)
                        )
                     )
                  ).
               expand_call(
                  some(used, callable(top, [])), F, Args, _, _, Goal
               ) :-
                  !,
                  Goal = callable(F, Args).
               expand_call(some(used, M), F, Args, Pb, Pe, Goal) :-
                  (
                     Fid = toplevel(Fid0)
                  ->
                     EmopCall = 'emop_call'
                  ;
                     EmopCall = 'emop_call',
                     Fid0 = Fid
                  ),
                  (
                     M = callable(Atom, [])
                  ->
                     ExpandFile:add_dependency(Atom, file(Pb, Pe))
                  ;
                     true
                  ),
                  'emop.message':position_text(Pb, Pe, Text),
                  EmopCallArgs = [M, callable(F, Args), callable(Text, [])],
                  Goal = callable(EmopCall, EmopCallArgs).
      
               expand_term(callable(F, _, Args0, _, _), Ctx, Goal, Term) :-
                  Expander:expand_term_list(Args0, Ctx, Goal, Args1),
                  Term = callable(F, Args1).
               expand_term(reserved(F, Args, Pb, Pe), Ctx, Goal, Term) :-
                  Expander:expand_term_reserved(
                     F, Args, Pb, Pe, Ctx, Goal, Term
                  ).
               expand_term('()'(SubTerm, _, _), Ctx, Goal, Term) :-
                  Expander:expand_term(SubTerm, Ctx, Goal, Term).
               expand_term('module'(Name, Env, Body, PB, PE), _, Goal, Term) :-
                  (
                     Name = var('_', _, _)
                  ->
                     ExpandFile:fresh_var(Term),
                     Term = var(FreshVarName),
                     Name0 = var(FreshVarName, PB, PE)
                  ;
                     Name = var(Var, _, _)
                  ->
                     Name0 = Name,
                     Term = var(Var)
                  ;
                     Name = atom(Atom, _, _, _)
                  ->
                     Name0 = Name,
                     Term = callable(Atom, [])
                  ), 
                  ExpandFile:expand_module(
                     Name0,
                     Env,
                     Body,
                     This,
                     Expander,
                     Goal
                  ).
               expand_term(number(Number, _, _), _, Goal, Term) :-
                  Goal = callable(true, []),
                  Term = number(Number).
               expand_term(var(Var, Pb, Pe), _, Goal, Term) :-
                  Goal = callable(true, []),
                  Expander:var_lookup(Var, Pb, Pe, _),
                  Term = var(Var).
               expand_term(empty(Pb, Pe), _, _, _) :-
                  throw(error('Empty term unallowed', file(Pb, Pe))).
               expand_term(doc(_, Pb, Pe, _), _, _, _) :-
                  throw(
                     error('Unallowed documentation here', file(Pb, Pe))
                  ).
      
               expand_term_reserved(':', [A0, B0], Pb, Pe, Ctx, Goal, Term) :-
                  !,
                  Expander:(
                     expand_term(A0, Ctx, AG, A1),
                     expand_subgoal(B0, some(Used, A1), BG, B1)
                  ),
                  Impl:check_prefix_used(Used, Pb, Pe),
                  Impl:seq(AG, BG, Goal),
                  Term = B1.
               expand_term_reserved('this', [], Pb, Pe, _, Goal, Term) :-
                  !,
                  Goal = callable(true, []),
                  (
                     This = none
                  ->
                     throw(
                        error(
                           '\'this\' cannot be used outside a module',
                           file(Pb, Pe)
                        )
                     )
                  ;
                     This = some(Term, _, _, _, _, _)
                  ).
               expand_term_reserved(F, Args0, _, _, Ctx, Goal, Term) :-
                  Expander:expand_term_list(Args0, Ctx, Goal, Args1),
                  Term = callable(F, Args1).
      
               expand_term_list([], _, callable(true, []), []).
               expand_term_list([H0 | T0], Ctx, G, [H1 | T1]) :-
                  Expander:(
                     expand_term(H0, Ctx, G0, H1),
                     expand_term_list(T0, Ctx, G1, T1)
                  ),
                  Impl:seq(G0, G1, G).
      
               expand_subgoal(callable(F, _, Args0, Pb, Pe), Ctx, Goal, Term) :-
                  Expander:expand_term_list(Args0, Ctx, G0, Args1),
                  ExpandFile:fresh_var(Term),
                  'data.list':append(Args1, [Term], Args2),
                  Expander:expand_call(Ctx, F, Args2, Pb, Pe, G1),
                  Impl:seq(G0, G1, Goal).
               expand_subgoal(reserved(_, _, Pb, Pe), _, _, _) :-
                  throw(
                     error(
                        'A predicate cannot use a reserved construction',
                        file(Pb, Pe)
                     )
                  ).
               expand_subgoal('()'(Subgoal, _, _), Ctx, Goal, Term) :-
                  Expander:expand_subgoal(Subgoal, Ctx, Goal, Term).
               expand_subgoal('module'(_, _, _, Pb, Pe), _, _, _) :-
                  throw(error('A module cannot be prefixed', file(Pb, Pe))).
               expand_subgoal(number(_, Pb, Pe), _, _, _) :-
                  throw(error('A number cannot be prefixed', file(Pb, Pe))).
               expand_subgoal(var(_, Pb, Pe), _, _, _) :-
                  throw(error('A variable cannot be prefixed', file(Pb, Pe))).
               expand_subgoal(empty(Pb, Pe), _, _, _) :-
                  throw(error('Empty goal unallowed', file(Pb, Pe))).
               expand_subgoal(doc(_, Pb, Pe, _), _, _, _) :-
                  throw(
                     error('Unallowed documentation here', file(Pb, Pe))
                  ).
            }.
      },
      Vars = 'data.record':{},
      ExpandFile:expander(Vars, none, Expander),
      Expander:expand_goal(Ast, none, Goal),
      DependenciesRef:get(Dependencies),
      ClausesRef:get(Clauses).
.
},

module Impl {
   extract_env(StructEnv, Expander, EnvRec, Env) :-
      module Extract [Extract, Expander, EnvRec] {
         do(callable(_, _, _, Pb, Pe), _) :-
            throw(
               error('A term cannot appear in an environment', file(Pb, Pe))
            ).
         do(reserved(_, _, Pb, Pe), _) :-
            throw(
               error(
                  'Environment components cannot use reserved construction',
                  file(Pb, Pe)
               )
            ).
         do('()'(SubEnv, _, _), Env) :-
            Extract:do(SubEnv, Env).
         do('module'(_, _, _, Pb, Pe), _) :-
            throw(
               error('A module cannot appear in an environment', file(Pb, Pe))
            ).
         do('number'(_,  Pb, Pe), _) :-
            throw(
               error('A number cannot appear in an environment', file(Pb, Pe))
            ).
         do('var'(Name, Pb, Pe), Result) :-
            (
               Name = '_'
            ->
               throw(
                  error('Anonymous environment variable', file(Pb, Pe))
               )
            ;
               true
            ),
            Var = EnvRec:{Name},
            (
               'data.var':check(Var)
            ->
               Expander:(
                  var_lookup(Name, Pb, Pe, _),
                  % Not singleton!
                  var_lookup(Name, _, _, _)
               ),
               Result = ['var'(Name)],
               Var = var(Pb, Pe, env, _)
            ;
               Result = [],
               'emop.message':warning(
                  'data.atom':format(
                     'Duplicated variable in environment \'~a\'',
                     [Name]
                  ),
                  file(Pb, Pe)
               )
            ).
         do('empty'(Pb, Pe), _) :-
            throw(
               error('Empty item unallowed in an environment', file(Pb, Pe))
            ).
         do(doc(_, Pb, Pe, _), _) :-
            throw(
               error('Unallowed documentation here', file(Pb, Pe))
            ).
      },
      'data.list':morph(StructEnv, Extract, Env).

   seq(callable(true, []), G, G) :-
      !.
   seq(G, callable(true, []), G) :-
      !.
   seq(G0, G1, callable(',', [G0, G1])).

   check_prefix_used(Used, Pb, Pe) :-
      (
         'data.var':check(Used)
      ->
         'emop.message':warning(
            'Unused prefix',
            file(Pb, Pe)
         )
      ;
         true
      ).
}
