module 'io.stream' [Impl] {
   backtrackable(Stream_nb, Stream_b) :-
      Impl:backtrackable(Stream_nb, Stream_b).

   text_position(Stream, File, Position) :-
      Impl:text_position(Stream, File, Position).

   list(Stream, List) :-
      Impl:list(Stream, List).
},

module Impl [Impl] {
   backtrackable(Stream_nb, Stream_b) :-
      'data.ref.non_backtrackable':(
         new([], Buffer),
         new(0, Current_char)
      ),
      module Stream_b [Stream_nb, Buffer, Current_char] {
         read(C) :-
            \+ \+ (
               Buffer:get(Buf),
               (
                  Buf = [H | T]
               ->
                  Buffer:set(T)
               ;
                  Stream_nb:read(H)
               ),
               Current_char:set(H)
            ),
            Current_char:get(H),
            (
               H = C
            ;
               Buffer:get(T),
               Buffer:set([H | T]),
               fail
            ). 
      }.

   text_position(Stream, File, Position) :-
      'data.ref.backtrackable':(
         new(0, Offset),
         new(0, Line),
         new(0, BOL)
      ),
      module Position [Stream, File, Offset, Line, BOL] {
         position(pos(File, O, L, B)) :-
            Offset:get(O),
            Line:get(L),
            BOL:get(B).

         read(C) :-
            Stream:read(C),
            'data.number':(Offset++),
            (
               C = '\n'
            ->
               Offset:get(O),
               BOL:set(O),
               'data.number':(Line++)
            ;
               true
            ).
      }.

   list(Stream, List) :-
      Stream:read(Item),
      (
         Item = end_of_file
      ->
         List = []
      ;
         List = [Item | Tail],
         Impl:list(Stream, Tail)
      ).
}
