module 'kernel' {
   array_spec([], InitialValue, InitialValue).
   array_spec([H | T], InitialValue, g_array(H, A)) :-
      array_spec(T, InitialValue, A).

   array_backtrackable_new(Sizes, InitialValue, Array) :-
      array_spec(Sizes, InitialValue, Spec),
      atom_new_backtrackable(Array),
      'top':g_link(Array, Spec).

   array_backtrackable_get(Array, Indexes, Value) :-
      'top':(
         '=..'(Term, [Array | Indexes]),
         g_read(Term, Value)
      ).

   array_backtrackable_set(Array, Indexes, Value) :-
      'top':(
         '=..'(Term, [Array | Indexes]),
         g_link(Term, Value)
      ).

   array_non_backtrackable_new(Sizes, InitialValue, Array) :-
      array_spec(Sizes, InitialValue, Spec),
      atom_new_backtrackable(Array),
      'top':g_assign(Array, Spec).

   array_non_backtrackable_get(Array, Indexes, Value) :-
      'top':(
         '=..'(Term, [Array | Indexes]),
         g_read(Term, Value)
      ).

   array_non_backtrackable_set(Array, Indexes, Value) :-
      'top':(
         '=..'(Term, [Array | Indexes]),
         g_assign(Term, Value)
      ).

   atom_check(Atom) :-
      'top':atom(Atom).

   atom_length(Atom, Length) :-
      'top':atom_length(Atom, Length).

   atom_new(Atom) :-
      'top':new_atom(Atom).

   atom_new_backtrackable(Atom) :-
      'top':(
         g_read(backtrackable_atom_count, AtomCount),
         g_read(backtrackable_atom_capacity, AtomCapacity),
         is(NewAtomCount, inc(AtomCount)),
         (
            NewAtomCount > AtomCapacity
         ->
            is(NewAtomCapacity, inc(AtomCapacity)),
            new_atom(Atom),
            g_assign(backtrackable_atoms(AtomCount), Atom),
            g_assign(backtrackable_atom_capacity, NewAtomCapacity)
         ;
            g_read(backtrackable_atoms(AtomCount), Atom)
         ),
         g_link(backtrackable_atom_count, NewAtomCount)
      ).

   atom_concat(A, B, R) :-
      'top':atom_concat(A, B, R).

   atom_format(Format, Args, Atom) :-
      'top':format_to_atom(Atom, Format, Args).

   atom_sub(Atom, Before, Length, After, SubAtom) :-
      'top':sub_atom(Atom, Before, Length, After, SubAtom).

   atom_to_string(Atom, String) :-
      'top':atom_chars(Atom, String).

   atom_to_number(Atom, Number) :-
      'top':number_atom(Number, Atom).

   char_to_code(Char, Code) :-
      'top':char_code(Char, Code).

   cmdline_count(Count) :-
      'top':argument_counter(Count).

   cmdline_get(Index, Argument) :-
      'top':argument_value(Index, Argument).

   cmdline_list(List) :-
      'top':argument_list(List).

   env_get(Key, Value) :-
      'top':environ(Key, Value).

   fd_check(X) :-
      'top':fd_var(X).

   fd_set_domain(X, Lower, Upper) :-
      'top':fd_domain(X, Lower, Upper).

   fd_set_domain(X, List) :-
      'top':fd_domain(X, List).

   fd_min(X, Min) :-
      'top':fd_min(X, Min).

   fd_max(X, Max) :-
      'top':fd_max(X, Max).

   fd_size(X, Size) :-
      'top':fd_size(X, Size).

   fd_dom(X, Dom) :-
      'top':fd_dom(X, Dom).

   fd_or(A, B, R) :-
      'top':'#<=>'(R, '#\\/'(A, B)).

   fd_and(A, B, R) :-
      'top':'#<=>'(R, '#/\\'(A, B)).

   fd_equiv(A, B, R) :-
      'top':'#<=>'(R, '#<=>'(A, B)).

   fd_equiv(A, B) :-
      'top':'#<=>'(A, B).

   fd_impl(A, B, R) :-
      'top':'#==>'(R, '#==>'(A, B)).

   fd_impl(A, B) :-
      'top':'#==>'(A, B).

   fd_cardinality(List, N) :-
      'top':fd_cardinality(List, N).

   fd_all_different(List) :-
      'top':fd_all_different(List).

   fd_element(I, List, X) :-
      'top':fd_element(I, List, X).

   fd_element_var(I, List, X) :-
      'top':fd_element_var(I, List, X).

   fd_labeling(X) :-
      'top':fd_labeling(X).

   fd_minimize(M) :-
      'top':fd_minimize(emop_call(M, objective(X), 'fd_minimize'), X).

   fd_maximize(M) :-
      'top':fd_maximize(emop_call(M, objective(X), 'fd_maximize'), X).

   fd_pac_opp(X, R) :-
      'top':'#='(R, '-'(X)).

   fd_pac_add(X, Y, R) :-
      'top':'#='(R, '+'(X, Y)).

   fd_pac_sub(X, Y, R) :-
      'top':'#='(R, '-'(X, Y)).

   fd_pac_mul(X, Y, R) :-
      'top':'#='(R, '*'(X, Y)).

   fd_pac_div(X, Y, R) :-
      'top':'#='(R, '/'(X, Y)).

   fd_pac_eq(A, B) :-
      'top':'#='(A, B).

   fd_pac_eq(A, B, R) :-
      'top':'#<=>'(R, '#='(A, B)).

   fd_pac_lt(A, B) :-
      'top':'#<'(A, B).

   fd_pac_lt(A, B, R) :-
      'top':'#<=>'(R, '#<'(A, B)).

   fd_pac_le(A, B) :-
      'top':'#=<'(A, B).

   fd_pac_le(A, B, R) :-
      'top':'#<=>'(R, '#=<'(A, B)).

   fd_pac_gt(A, B) :-
      'top':'#>'(A, B).

   fd_pac_gt(A, B, R) :-
      'top':'#<=>'(R, '#>'(A, B)).

   fd_pac_ge(A, B) :-
      'top':'#>='(A, B).

   fd_pac_ge(A, B, R) :-
      'top':'#<=>'(R, '#>='(A, B)).

   fd_fac_opp(X, R) :-
      'top':'#=#'(R, '-'(X)).

   fd_fac_add(X, Y, R) :-
      'top':'#=#'(R, '+'(X, Y)).

   fd_fac_sub(X, Y, R) :-
      'top':'#=#'(R, '-'(X, Y)).

   fd_fac_mul(X, Y, R) :-
      'top':'#=#'(R, '*'(X, Y)).

   fd_fac_div(X, Y, R) :-
      'top':'#=#'(R, '/'(X, Y)).

   fd_fac_eq(A, B) :-
      'top':'#=#'(A, B).

   fd_fac_eq(A, B, R) :-
      'top':'#<=>'(R, '#=#'(A, B)).

   fd_fac_lt(A, B) :-
      'top':'#<#'(A, B).

   fd_fac_lt(A, B, R) :-
      'top':'#<=>'(R, '#<#'(A, B)).

   fd_fac_le(A, B) :-
      'top':'#=<#'(A, B).

   fd_fac_le(A, B, R) :-
      'top':'#<=>'(R, '#=<#'(A, B)).

   fd_fac_gt(A, B) :-
      'top':'#>#'(A, B).

   fd_fac_gt(A, B, R) :-
      'top':'#<=>'(R, '#>#'(A, B)).

   fd_fac_ge(A, B) :-
      'top':'#>=#'(A, B).

   fd_fac_ge(A, B, R) :-
      'top':'#<=>'(R, '#>=#'(A, B)).

   io_open_read(Filename, Stream) :-
      'top':open(Filename, read, Stream).

   io_open_write(Filename, Stream) :-
      'top':open(Filename, write, Stream).

   io_read(Stream, Char) :-
      'top':get_char(Stream, Char).

   io_write(Stream, Char) :-
      'top':put_char(Stream, Char).

   io_flush(Stream) :-
      'top':flush_output(Stream).

   io_print(Stream, Term) :-
      'top':print(Stream, Term).

   io_print_endline(Stream, Term) :-
      'top':(
         print(Stream, Term),
         nl(Stream)
      ).

   io_format(Stream, Format, Args) :-
      'top':format(Stream, Format, Args).

   io_close(Stream) :-
      'top':close(Stream).

   list_length(L, N) :-
      'top':length(L, N).

   number_check(X) :-
      'top':number(X).

   number_char(Code, Char) :-
      'top':char_code(Char, Code).

   number_ge(A, B) :-
      'top':'>='(A, B)
.
   number_gt(A, B) :-
      'top':'>'(A, B).

   number_le(A, B) :-
      'top':'=<'(A, B)
.
   number_lt(A, B) :-
      'top':'<'(A, B).

   number_pred(X, Y) :-
      'top':is(Y, dec(X)).

   number_succ(X, Y) :-
      'top':is(Y, inc(X)).

   number_opp(X, R) :-
      'top':is(R, '-'(X)).

   number_add(X, Y, R) :-
      'top':is(R, '+'(X, Y)).

   number_sub(X, Y, R) :-
      'top':is(R, '-'(X, Y)).

   number_mul(X, Y, R) :-
      'top':is(R, '*'(X, Y)).

   number_div(X, Y, R) :-
      'top':is(R, '/'(X, Y)).

   number_quo(X, Y, R) :-
      'top':is(R, '//'(X, Y)).

   number_mod(X, Y, R) :-
      'top':is(R, 'mod'(X, Y)).

   number_to_atom(N, A) :-
      'top':number_atom(N, A).

   ref_backtrackable_new(Initial, Ref) :-
      atom_new_backtrackable(Ref),
      'top':g_link(Ref, Initial).

   ref_backtrackable_get(Ref, Value) :-
      'top':g_read(Ref, Value).

   ref_backtrackable_set(Ref, Value) :-
      'top':g_link(Ref, Value).

   ref_non_backtrackable_new(Initial, Ref) :-
      atom_new_backtrackable(Ref),
      'top':g_assign(Ref, Initial).

   ref_non_backtrackable_get(Ref, Value) :-
      'top':g_read(Ref, Value).

   ref_non_backtrackable_set(Ref, Value) :-
      'top':g_assign(Ref, Value).

   stderr_endline :-
      'top':(
         g_read(stderr, Stderr),
         nl(Stderr)
      ).

   stderr_flush :-
      'top':(
         g_read(stderr, Stderr),
         'top':flush_output(Stderr)
      ).

   stderr_format(Format, Args) :-
      'top':(
         g_read(stderr, Stderr),
         format(Stderr, Format, Args)
      ).

   stderr_print(Term) :-
      'top':(
         g_read(stderr, Stderr),
         print(Stderr, Term)
      ).

   stderr_print_endline(Term) :-
      'top':(
         g_read(stderr, Stderr),
         print(Stderr, Term),
         nl(Stderr)
      ).

   stdio_read(C) :-
      'top':get_char(C).

   stdio_write(C) :-
      'top':put_char(C).

   stdio_endline :-
      'top':nl.

   stdio_flush :-
      'top':flush_output.

   stdio_format(Format, Args) :-
      'top':format(Format, Args).

   stdio_print(Term) :-
      'top':print(Term).

   stdio_print_endline(Term) :-
      'top':(
         print(Term),
         nl
      ).

   string_to_atom(String, Atom) :-
      'top':atom_chars(Atom, String).

   string_to_number(String, Number) :-
      'top':number_chars(Number, String).

   sys_abort :-
      'top':abort.

   sys_exit(Error_code) :-
      'top':halt(Error_code).

   sys_halt :-
      'top':halt.

   sys_spawn(Command, Arguments, Error_code) :-
      'top':spawn(Command, Arguments, Error_code).

   sys_file_real(File, Real) :-
      'top':file_property(File, real_file_name(Real)).

   sys_file_absolute(File, Absolute) :-
      'top':file_property(File, absolute_file_name(Absolute)).

   random_randomize :-
      'top':randomize.

   random_get_seed(Seed) :-
      'top':get_seed(Seed).

   random_set_seed(Seed) :-
      'top':set_seed(Seed).

   random_random(Float) :-
      'top':random(Float).

   random_random(Base, Max, Int) :-
      'top':random(Base, Max, Int).

   record_new(X) :-
      atom_new_backtrackable(X),
      'top':g_link(X, []).

   record_get(X, Key, Value) :-
      'top':(
         atom_concat(X, Key, Var),
         (
            g_read(Var, some(Value))
         ->
            true
         ;
            g_link(Var, some(Value)),
            g_read(X, Keys),
            g_link(X, [Key | Keys])
         )
      ).

   record_keys(X, List) :-
      'top':g_read(X, List).

   var_check(Var) :-
      'top':generic_var(Var).

   var_pure(Var) :-
      'top':var(Var).

   term_check(X) :-
      'top':callable(X).

   term_compare(A, B, R) :-
      'top':compare(R, A, B).

   term_ge(A, B) :-
      'top':'@>='(A, B)
.
   term_gt(A, B) :-
      'top':'@>'(A, B).

   term_le(A, B) :-
      'top':'@=<'(A, B)
.
   term_lt(A, B) :-
      'top':'@<'(A, B).

   term_decomposition(Term, List) :-
      'top':'=..'(Term, List).

   term_functor(Term, F, N) :-
      'top':functor(Term, F, N).
},
'top':(
   open('/dev/fd/2', write, Stderr),
   g_assign(stderr, Stderr),
   g_assign(backtrackable_atoms, g_array_auto(7, 0)),
   g_assign(backtrackable_atom_capacity, 0),
   g_link(backtrackable_atom_count, 0)
)
